/* vi:set ts=4 sw=4:

  textbuf.c    gapped buffer for ruby

*/

#include <stdio.h>
#ifdef __STDC__
# include <stdlib.h>
#endif

#include "ruby.h"
#include "re.h"


/* declear ------------------------------------------------------ */

/*
    === struct textbuf ===

                gap/cgap
                   |
                   V
    ptr -> OoOoOoOo--------------OoOoOoOoOoOoOo
                   |<- gaplen ->|
           |<------------ len --------------->|

    O : first byte of multi byte char
    o : second byte
    - : gap


    *  gap never divide multi byte char

*/

struct textbuf {
    char *ptr;              /* buffer */
    long len;               /* byte length of buffer (gap length added) */
    long clen;              /* char length of buffer (gap length not added) */
    long capa;              /* byte capacity of buffer */
    long gap;               /* byte index of gap */
    long cgap;              /* char index of gap */
    long gaplen;            /* byte length of gap space */

    struct bufmark *marks;  /* d-linked-list of mark */
    VALUE cursor;           /* cursor mark */
    struct bufmark *point;  /* edit point (internal use only) */
};

struct bufmark {
    unsigned long flags;
    long idx;
    long cidx;
    struct textbuf *buf;
    struct bufmark *prev;
    struct bufmark *next;
};


static VALUE TextBuffer;
static VALUE TextBufferMark;


#define GET_BUF(v,var) Data_Get_Struct(v, struct textbuf, var)
#define GET_MARK(v, var) Data_Get_Struct(v, struct bufmark, var)

#define WRAP_BUF(p) Data_Wrap_Struct(TextBuffer, textbuf_mark, textbuf_free, p)
#define WRAP_MARK(p) Data_Wrap_Struct(TextBufferMark, 0, bufmark_free, p)


#define ALIVE_FLAG    (1U << 0)  /* mark->buf is not freed */
#define INSERT_BEFORE (1U << 1)  /* insert text in "left" of mark */

#define ALIVE_P(m) (m->flags & ALIVE_FLAG)

#define INITLEN 1024


/* util -------------------------------------------------------- */

static void
p_buf(buf)
    struct textbuf *buf;
{
    printf("buf.ptr=0x%lx, buf.len=%ld, buf.clen=%ld\n",
           (unsigned long)buf->ptr, buf->len, buf->clen);
}

static void
p_mark(mark)
    struct bufmark *mark;
{
    printf("mark.buf=0x%lx mark.idx=%ld mark.cidx=%ld\n",
           (unsigned long)mark->buf, mark->idx, mark->cidx);
}


static long
legal_cidx(buf, c)
    struct textbuf *buf;
    long c;
{
    if (c < 0) return 0;
    if (c > buf->clen) return buf->clen;
    return  c;
}


static long
legal_idx(buf, i)
    struct textbuf *buf;
    long i;
{
    if (i < 0) return 0;
    if (i > buf->len) return buf->len;
    return  i;
}


static long
byte_length(buf, begidx, clen)
    struct textbuf *buf;
    long begidx, clen;
{
    long i = begidx;
    long c = clen;

    while (c && i < buf->len) {
        i += mbclen(buf->ptr[i]);
        c--;
    }
/* printf("byte_len: beg=%ld, clen=%ld, len=%ld\n", begidx, clen, i - begidx);
*/
    return i - begidx;
}


static long
char_length(buf, begidx, len)
    struct textbuf *buf;
    long begidx, len;
{
    long i = begidx;
    long c = 0;
    long targ;

    targ = legal_idx(buf, begidx + len);
    while (i < targ) {
        i += mbclen(buf->ptr[i]);
        c++;
    }
printf("char_len: beg=%ld, len=%ld, clen=%ld\n", begidx, len, c);
    return c;
}


static void
alive_check(mark)
    struct bufmark *mark;
{
    if (ALIVE_P(mark))
        return;
    else
        rb_raise(rb_eArgError, "method called for dead mark");
}



/* mark list handling ------------------------------------------ */

static void
insert_mark_between(mark, prev, next)
    struct bufmark *mark, *prev, *next;
{
    mark->prev = prev;
    mark->next = next;
    if (prev) prev->next = mark;
    if (next) next->prev = mark;
}

static void
remove_mark(mark)
    struct bufmark *mark;
{
    if (mark->prev)
        mark->prev->next = mark->next;
    if (mark->next)
        mark->next->prev = mark->prev;
    mark->prev = 0;
    mark->next = 0;
}

static struct bufmark*
first_mark(mark)
    struct bufmark *mark;
{
    struct bufmark *top;

    top = mark->buf->marks ? mark->buf->marks : mark;
    while (top->prev)
        top = top->prev;
    
    return top;
}

static void
adjust_mark_forward(mark)
    struct bufmark *mark;
{
    struct bufmark *prev, *next;

    prev = mark->prev;
    next = mark->next;
    remove_mark(mark);

    while (next) {
        if (mark->idx == next->idx) {
            if ((mark->flags & INSERT_BEFORE) <=
                (next->flags & INSERT_BEFORE))
                /* ins-after is "little" than ins-before */
                break;
        }
        else if (mark->idx < next->idx) {
            /* if ((prev && prev->idx < mark->idx) || !prev) */
            break;
        }
        prev = next;
        next = next->next;
    }
    insert_mark_between(mark, prev, next);

    mark->buf->marks = first_mark(mark);
}

static void
adjust_mark_backward(mark)
    struct bufmark *mark;
{
    struct bufmark *top, *prev, *next;

    prev = mark->prev;
    next = mark->next;
    remove_mark(mark);

    while (prev) {
        if (prev->idx == mark->idx) {
            if ((prev->flags & INSERT_BEFORE) <=
                (mark->flags & INSERT_BEFORE))
                /* ins-after is "little" than ins-before */
                break;
        }
        else if (prev->idx < mark->idx) {
            /* if ((next && mark->idx < next->idx) || !next) */
            break;
        }
        next = prev;
        prev = prev->prev;
    }

    insert_mark_between(mark, prev, next);

    top = mark->buf->marks ? mark->buf->marks : mark;
    while (top->prev)
        top = top->prev;
    mark->buf->marks = top;
}



/* creation -------------------------------------------------------- */

static void
bufmark_free(mark)
    struct bufmark *mark;
{
    if (ALIVE_P(mark))
        remove_mark(mark);
    free(mark);
}

static int
bufmark_p(obj)
    VALUE obj;
{
    if (TYPE(obj) == T_DATA)
        if (RDATA(obj)->dfree == bufmark_free)
            return 1;
    return 0;
}

static struct bufmark*
new_mark_intern(buf, idx, cidx)
    struct textbuf *buf;
    long idx, cidx;
{
    struct bufmark *mark;

    mark = ALLOC(struct bufmark);
    mark->buf  = buf;
    mark->idx  = idx;
    mark->cidx = cidx;
    mark->prev = 0;
    mark->next = 0;

    mark->flags = 0;
    mark->flags |= ALIVE_FLAG | INSERT_BEFORE;

    mark->next = buf->marks;
    buf->marks = mark;
    adjust_mark_forward(mark);

    return mark;
}


static void
textbuf_mark(buf)
    struct textbuf *buf;
{
    rb_gc_mark(buf->cursor);
}

static void
textbuf_free(buf)
    struct textbuf *buf;
{
    struct bufmark *m;

    free(buf->ptr);
    buf->ptr = 0;
    buf->len = 0;
    buf->clen = 0;
    buf->gap = 0;
    buf->cgap = 0;

    buf->cursor = Qnil;
    remove_mark(buf->point);
    free(buf->point);
    buf->point = 0;
    for (m = buf->marks; m; m = m->next)
        m->flags &= ~ALIVE_FLAG;
    buf->marks = 0;
    
    free(buf);
}

static VALUE
textbuf_s_new(self)
    VALUE self;
{
    struct textbuf *buf;

    buf = ALLOC(struct textbuf);
    buf->ptr = 0;
    buf->len = 0;
    buf->clen = 0;
    buf->capa = 0;
    buf->gap = 0;
    buf->cgap = 0;
    buf->cursor = Qnil;
    buf->point = 0;

    buf->ptr = ALLOC_N(char, INITLEN);
    buf->capa = INITLEN;

    buf->cursor = WRAP_MARK(new_mark_intern(buf, 0, 0));
    buf->point  = new_mark_intern(buf, 0, 0);

    return WRAP_BUF(buf);
}


static VALUE
textbuf_new_mark(self, cidx)
    VALUE self, cidx;
{
    struct textbuf *buf;
    struct bufmark *mark;
    long c, i;

    GET_BUF(self, buf);
    c = legal_cidx(buf, NUM2LONG(cidx));
    i = char_length(buf, 0, c);
    mark = new_mark_intern(buf, i, c);

    return WRAP_MARK(mark);
}



/* move mark ------------------------------------------------------ */

static void
forward_mark_intern(mark, clen)
    struct bufmark *mark;
    long clen;
{
    struct bufmark *m;
    long targ;

    targ = legal_cidx(mark->buf, mark->cidx + clen);
    for (m = mark; m->next; m = m->next) {
        if (m->next->cidx > targ)
            break;
    }
    mark->idx = m->idx + byte_length(mark->buf, m->idx, targ - m->cidx);
    mark->cidx = targ;

    adjust_mark_forward(mark);
}

static void
back_mark_intern(mark, clen)
    struct bufmark *mark;
    long clen;
{
    struct bufmark *m;
    long targ, i, cl;

    targ = legal_cidx(mark->buf, mark->cidx - clen);
    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->cidx < targ) {
            m = m->prev;
            break;
        }
    }
    if (m->cidx < targ) {
        i = m->idx;
        cl = targ - m->cidx;
    }
    else {
        i = 0;
        cl = targ;
    }
    mark->idx = i + byte_length(mark->buf, i, cl);
    mark->cidx = targ;

    adjust_mark_backward(mark);
}



/* buffer & mark handling functions --------------------------------- */

static void
move_mark_for_insert(mark, len, clen)
    struct bufmark *mark;
    long len, clen;
{
    struct bufmark *m;

    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != m->idx)
            break;
    }
    for (; m && m->idx == mark->idx; m = m->next) {
        if (m->flags & INSERT_BEFORE) {
            m->idx += len;
            m->cidx += clen;
        }
    }
    for (; m; m = m->next) {
        m->idx += len;
        m->cidx += clen;
    }
}

static void
insert_intern(mark, str, len)
    struct bufmark *mark;
    char *str;
    long len;
{
    long clen;
    long idx = mark->idx;
    struct textbuf *buf = mark->buf;

    if (buf->capa < buf->len + len) {
        long newcapa = buf->capa;
        while (newcapa < buf->len + len)
            newcapa *= 2;
        REALLOC_N(buf->ptr, char, newcapa);
    }

    memmove(buf->ptr + idx + len,
            buf->ptr + idx,
            (buf->len - idx) * sizeof(char));
    memmove(buf->ptr + idx,
            str,
            sizeof(char) * len);
    buf->len += len;
    clen = char_length(buf, idx, len);
    buf->clen += clen;

    move_mark_for_insert(mark, len, clen);
}


static void
move_mark_for_delete(mark, len, clen)
    struct bufmark *mark;
    long len, clen;
{
    struct bufmark *m;
    long upper;

    upper = mark->idx + len;

    for (m = mark; m->prev; m = m->prev) {
        if (m->prev->idx != m->idx)
            break;
    }
    for (; m && m->idx < upper; m = m->next) {
        m->idx = mark->idx;
        m->cidx = mark->cidx;
    }
    for (; m; m = m->next) {
        m->idx -= len;
        m->cidx -= clen;
    }
}

static void
delete_intern(mark, len)
    struct bufmark *mark;
    long len;
{
    long clen;
    long idx = mark->idx;
    struct textbuf *buf = mark->buf;

printf("\ndelete_intern\n");
p_mark(mark);
p_buf(buf);
    memmove(buf->ptr + idx,
            buf->ptr + (idx + len),
            buf->len - (idx + len));
    clen = char_length(buf, idx, len);
    buf->len -= len;
    buf->clen -= clen;

    move_mark_for_delete(mark, len, clen);
}


static VALUE
substr_intern(buf, idx, len)
    struct textbuf *buf;
    long idx, len;
{
    return rb_str_new(buf->ptr + idx, len);

/*
    if (cidx > buf->cgap || cidx + clen < buf->cgap) {
        return rb_str_new(cidx2idx(cidx), byte_length(buf, cidx, clen));
    }
    else {
    }
*/
}


static VALUE
extract_intern(buf, idx, len)
    struct textbuf *buf;
    long idx, len;
{
    NEWOBJ(str, struct RString);
    OBJSETUP(str, rb_cString, T_STRING);
    str->ptr = buf->ptr + idx;
    str->len = len;
    str->orig = 0;
    OBJ_FREEZE(str);

    if (rb_iterator_p()) {
        rb_yield((VALUE)str);
        str->ptr = 0;
        str->len = 0;
        return Qnil;
    }
    else {
        return (VALUE)str;
    }
}



/* buffer interfaces ---------------------------------------------- */

static VALUE
textbuf_cursor(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return buf->cursor;
}


static VALUE
textbuf_size(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return INT2NUM(buf->clen);
}


static VALUE
textbuf_byte_size(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return INT2NUM(buf->len);
}


static VALUE
textbuf_substr(self, cidx, clen)
    VALUE self, cidx, clen;
{
    struct textbuf *buf;
    long ci, cl;

    GET_BUF(self, buf);
    ci = legal_cidx(buf, NUM2LONG(cidx));
    cl = NUM2LONG(clen);
    cl = legal_cidx(buf, ci + cl) - ci;
    if (buf->point->cidx > ci)
        back_mark_intern(buf->point, buf->point->cidx - ci);
    else
        forward_mark_intern(buf->point, ci - buf->point->cidx);

    return substr_intern(buf, buf->point->cidx, cl);
}


static VALUE
textbuf_inspect(self)
    VALUE self;
{
    struct textbuf *buf;
    char tmp[1024];
    VALUE ret;
    int len;

    GET_BUF(self, buf);
    ret = rb_str_new("#<TextBuffer: ", 14);
    len = sprintf(tmp, "0x%lx ", (unsigned long)buf);
    rb_str_cat(ret, tmp, len);
    len = sprintf(tmp, "len=%ld ", buf->len);
    rb_str_cat(ret, tmp, len);
    len = sprintf(tmp, "clen=%ld ", buf->clen);
    rb_str_cat(ret, tmp, len);
    len = sprintf(tmp, "gap=%ld ", buf->gap);
    rb_str_cat(ret, tmp, len);
    len = sprintf(tmp, "cgap=%ld", buf->cgap);
    rb_str_cat(ret, tmp, len);
    rb_str_cat(ret, ">", 1);

    return ret;
}


static VALUE
textbuf_string(self)
    VALUE self;
{
    struct textbuf *buf;

    GET_BUF(self, buf);
    return rb_str_new(buf->ptr, buf->len);
}



/* mark interfaces -------------------------------------------- */

static VALUE
bufmark_dup(self)
    VALUE self;
{
    struct bufmark *mark, *p;

    GET_MARK(self, mark);
    alive_check(mark);
    p = new_mark_intern(mark->buf, mark->idx, mark->cidx);
    
    return WRAP_MARK(p);
}


static VALUE
bufmark_alive_p(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    return ALIVE_P(mark) ? Qtrue : Qfalse;
}


static VALUE
bufmark_insert(self, str)
    VALUE self, str;
{
    struct bufmark *mark;
    
    GET_MARK(self, mark);
    alive_check(mark);
    insert_intern(mark, RSTRING(str)->ptr, RSTRING(str)->len);

    return Qnil;
}

static VALUE
bufmark_delete(self, len)
    VALUE self, len;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    alive_check(mark);
    delete_intern(mark, NUM2LONG(len));

    return Qnil;
}


static VALUE
bufmark_substr(self, clen)
    VALUE self, clen;
{
    struct bufmark *mark;
    long len;

    GET_MARK(self, mark);
    alive_check(mark);
    len = byte_length(mark->buf, mark->idx, NUM2LONG(clen));

    return substr_intern(mark->buf, mark->idx, len);
}

static VALUE
bufmark_extract(self, clen)
    VALUE self, clen;
{
    struct bufmark *mark;
    long len;

    GET_MARK(self, mark);
    alive_check(mark);
    len = byte_length(mark->buf, mark->idx, NUM2LONG(clen));

    return extract_intern(mark->buf, mark->idx, len);
}


static VALUE
bufmark_fwd(self, clen)
    VALUE self, clen;
{
    struct bufmark *mark;
    long c;

    GET_MARK(self, mark);
    c = NUM2LONG(clen);
    if (c >= 0)
        forward_mark_intern(mark, c);
    else
        back_mark_intern(mark, -c);

    return self;
}

static VALUE
bufmark_back(self, clen)
    VALUE self, clen;
{
    struct bufmark *mark;
    long c;

    GET_MARK(self, mark);
    c = NUM2LONG(clen);
    if (c >= 0)
        back_mark_intern(mark, c);
    else
        forward_mark_intern(mark, -c);

    return self;
}

static VALUE
bufmark_move(self, clen)
    VALUE self, clen;
{
    struct bufmark *mark;
    long c;

    GET_MARK(self, mark);
    c = NUM2LONG(clen);
    if (c >= 0)
        forward_mark_intern(mark, c);
    else
        back_mark_intern(mark, -c);
    
    return self;
}


static VALUE
bufmark_getidx(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    alive_check(mark);
    return INT2NUM(mark->cidx);
}

static VALUE
bufmark_setidx(self, cidx)
    VALUE self;
{
    struct bufmark *mark;
    long c = 0;

    if (TYPE(cidx) == T_FIXNUM) {
        c = FIX2LONG(cidx);
    }
    else if (TYPE(cidx) == T_BIGNUM) {
        rb_raise(rb_eArgError, "too big index");
    }
    else if (bufmark_p(cidx)) {
        struct bufmark *m;
        GET_MARK(cidx, m);
        c = m->cidx;
    }
    else {
        rb_raise(rb_eTypeError, "wrong argument type %s",
                 rb_class2name(CLASS_OF(cidx)));
    }
    GET_MARK(self, mark);
    alive_check(mark);
    if (c > mark->cidx)
        forward_mark_intern(mark, c - mark->cidx);
    else
        back_mark_intern(mark, mark->cidx - c);

    return self;
}


static VALUE
bufmark_byte_index(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    alive_check(mark);
    return INT2NUM(mark->idx);
}


static VALUE
bufmark_go_beg(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    alive_check(mark);

    mark->idx = 0;
    mark->cidx = 0;
    adjust_mark_backward(mark);

    return self;
}

static VALUE
bufmark_go_end(self)
    VALUE self;
{
    struct bufmark *mark;

    GET_MARK(self, mark);
    alive_check(mark);

    mark->idx = mark->buf->len;
    mark->cidx = mark->buf->clen;
    adjust_mark_forward(mark);

    return self;
}


static VALUE
bufmark_inspect(self)
    VALUE self;
{
    struct bufmark *mark;
    VALUE ret;
    char buf[1024];
    int len;

    GET_MARK(self, mark);
    ret = rb_str_new("#<TextBufferMark: ", 18);
    len = sprintf(buf, "0x%lx ", (unsigned long)mark);
    rb_str_cat(ret, buf, len);
    len = sprintf(buf, "buf=0x%lx ", (unsigned long)mark->buf);
    rb_str_cat(ret, buf, len);
    len = sprintf(buf, "i=%ld ", mark->idx);
    rb_str_cat(ret, buf, len);
    len = sprintf(buf, "ci=%ld", mark->cidx);
    rb_str_cat(ret, buf, len);
    rb_str_cat(ret, ">", 1);

    return ret;
}



/* ------------------------------------------------------------- */

void
Init_textbuf()
{
    TextBuffer = rb_define_class("TextBuffer", rb_cData);
    rb_define_singleton_method(TextBuffer, "new", textbuf_s_new,   0);
    rb_define_method(TextBuffer, "new_mark",    textbuf_new_mark,  1);
    rb_define_method(TextBuffer, "cursor",      textbuf_cursor,    0);
    rb_define_method(TextBuffer, "size",        textbuf_size,      0);
    rb_define_method(TextBuffer, "length",      textbuf_size,      0);
    rb_define_method(TextBuffer, "byte_size",   textbuf_byte_size, 0);
    rb_define_method(TextBuffer, "byte_length", textbuf_byte_size, 0);
    rb_define_method(TextBuffer, "[]",          textbuf_substr,    2);
    rb_define_method(TextBuffer, "inspect",     textbuf_inspect,   0);
    rb_define_method(TextBuffer, "string",      textbuf_string,    0);

    TextBufferMark = rb_define_class("TextBufferMark", rb_cData);
    rb_define_method(TextBufferMark, "dup",        bufmark_dup,        0);
    rb_define_method(TextBufferMark, "alive?",     bufmark_alive_p,    0);
    rb_define_method(TextBufferMark, "insert",     bufmark_insert,     1);
    rb_define_method(TextBufferMark, "delete",     bufmark_delete,     1);
    rb_define_method(TextBufferMark, "substr",     bufmark_substr,     1);
    rb_define_private_method(TextBufferMark, "extract", bufmark_extract, 1);
    rb_define_method(TextBufferMark, "[]",         bufmark_substr,     1);
    rb_define_method(TextBufferMark, "index",      bufmark_getidx,     0);
    rb_define_method(TextBufferMark, "index=",     bufmark_setidx,     0);
    rb_define_method(TextBufferMark, "move_to",    bufmark_setidx,     0);
    rb_define_method(TextBufferMark, "byte_index", bufmark_byte_index, 0);
    rb_define_method(TextBufferMark, "move",       bufmark_move,       1);
    rb_define_method(TextBufferMark, "forward",    bufmark_fwd,        1);
    rb_define_method(TextBufferMark, "back",       bufmark_back,       1);
    rb_define_method(TextBufferMark, "go_head",    bufmark_go_beg,     0);
    rb_define_method(TextBufferMark, "go_tail",    bufmark_go_end,     0);
    rb_define_method(TextBufferMark, "inspect",    bufmark_inspect,    0);
}
